from sklearn.feature_extraction.text import TfidfVectorizer
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
import string
import numpy as np

def vectorFromTFIDF(df, column_name) -> list:          
        column_data = df[column_name].tolist()
        if not isinstance(column_data[0], str):
            column_data = [str(x) for x in column_data]
        column_data = [item for item in column_data if len(item)!=0]
        if len(column_data) == 0:
            return None        

        def custom_tokenizer(text):
            # Tokenize words
            if isinstance(text,str):
                tokens = word_tokenize(text.lower())
            # Remove stopwords
            stop_words = set(stopwords.words('english') + list(string.punctuation))
            filtered_tokens = [token for token in tokens if token not in stop_words]
            # Apply stemming or lemmatization
            stemmer = PorterStemmer()
            # You can choose either stemming or lemmatization here
            # For this example, we'll use lemmatization
            processed_tokens = [stemmer.stem(token) for token in filtered_tokens]
            if len(processed_tokens) == 0:
                return None
            return processed_tokens

        tokenized_data = [custom_tokenizer(x) for x in column_data]
        tokenized_data = [item for item in tokenized_data if item is not None]
            
        if len(tokenized_data) == 0:
                tokenized_data = column_data    
        def preprocessor(t):
            return t
        def tokenizer(t):
            return t 
        # Create TfidfVectorizer object
        vectorizer = TfidfVectorizer(max_features=1000, preprocessor=preprocessor, tokenizer=tokenizer, stop_words=None)
        tfidf_matrix = vectorizer.fit_transform( tokenized_data)
        # Get feature names (words)
        feature_names = vectorizer.get_feature_names_out()
        # Calculate TF-IDF scores for each word
        scores = tfidf_matrix.toarray().sum(axis=0)
        # Sort words by score and select top 512
        top_512_words = np.argsort(scores)[::-1][:512]
        top_512_words_list = [feature_names[i] for i in top_512_words]
        # top_512_words_list.append(column_name)
        return top_512_words_list


import pandas as pd
from pandas import DataFrame
from ..datamodel import Data
from .utils import vectorFromTFIDF

class Schema(Data): 
    
    
    """
        The corpus of the dataset that will be processed with pyjedai. \
        Contains all the information of the dataset and will be passed to each step \
        of the ER workflow. 
    """

    def __init__(
                self,
                dataset_1: DataFrame,
                dataset_2: DataFrame,
                dataset_name_1: str = None,
                dataset_name_2: str = None,
                ground_truth: DataFrame = None,
                skip_ground_truth_processing: bool = False,
                matching_type: str = None,
    ) -> None:
        
        # Original Datasets as pd.DataFrame
        if not isinstance(dataset_1, pd.DataFrame):
            raise AttributeError("Dataset 1 must be a pandas DataFrame")

        if not isinstance(dataset_2, pd.DataFrame):
            raise AttributeError("Dataset 2 must be a pandas DataFrame")
                    
        if matching_type == 'CONTENT':
            dataset_1, dataset_2, ground_truth = self.load_content(dataset_1, dataset_2, ground_truth, skip_ground_truth_processing)
        elif matching_type == 'COMPOSITE': 
            dataset_1, dataset_2, ground_truth = self.load_composite(dataset_1, dataset_2, ground_truth, skip_ground_truth_processing)
        else:
            dataset_1, dataset_2, ground_truth = self.load_schema(dataset_1, dataset_2, ground_truth, skip_ground_truth_processing) 
            
        super().__init__(dataset_1 = dataset_1, 
                    id_column_name_1 = 'id',
                    dataset_name_1 = dataset_name_1,
                    dataset_2 = dataset_2, 
                    id_column_name_2 = 'id',
                    dataset_name_2 = dataset_name_2,
                    ground_truth = ground_truth,
                    skip_ground_truth_processing = skip_ground_truth_processing)
        


    def load_content(self,
                dataset_1: DataFrame, 
                dataset_2: DataFrame,
                ground_truth: DataFrame = None,
                skip_ground_truth_processing: bool = False) -> tuple:
        
        dataset_1 = dataset_1.astype(str)
        dataset_2 = dataset_2.astype(str)

        source_attributes = dataset_1.columns
        target_attributes = dataset_2.columns

        source_index = range(len(source_attributes)) 
        source_data = [vectorFromTFIDF(dataset_1, col) for col in source_attributes]

        target_index = range(len(target_attributes)) 
        target_data = [vectorFromTFIDF(dataset_2, col) for col in target_attributes]


        dataset_1 = pd.DataFrame({
            'id' : source_index,
            'data': source_data
        })

        dataset_2 = pd.DataFrame({
            'id' : target_index,
            'data': target_data
        })


        dataset_1_columns = pd.DataFrame({
            'source': source_attributes,
            "source_index": source_index
        })

        dataset_2_columns = pd.DataFrame({ 
            "target" : target_attributes,
            "target_index": target_index
        })


        if ground_truth is not None and not skip_ground_truth_processing:
            ground_truth.columns = ['source', 'target']
            ground_truth = pd.merge(ground_truth, dataset_1_columns, on="source", how='left')
            ground_truth = pd.merge(ground_truth, dataset_2_columns, on='target', how='left')
            ground_truth = ground_truth.drop(columns=['source', 'target'])

        return dataset_1, dataset_2, ground_truth
    
    def load_composite(self,
            dataset_1: DataFrame, 
            dataset_2: DataFrame,
            ground_truth: DataFrame = None,
            skip_ground_truth_processing: bool = False
    ) -> tuple:
        dataset_1 = dataset_1.astype(str)
        dataset_2 = dataset_2.astype(str)

        source_attributes = dataset_1.columns
        target_attributes = dataset_2.columns

        source_index = range(len(source_attributes)) 
        source_data = [vectorFromTFIDF(dataset_1, col) for col in source_attributes]

        target_index = range(len(target_attributes)) 
        target_data = [vectorFromTFIDF(dataset_2, col) for col in target_attributes]


        dataset_1 = pd.DataFrame({
            'id' : source_index,
            'attributes': source_attributes,
            'data': source_data
        })

        dataset_2 = pd.DataFrame({
            'id' : target_index,
            'attributes': target_attributes,
            'data': target_data
        })


        dataset_1_columns = pd.DataFrame({
            'source': source_attributes,
            "source_index": source_index
        })

        dataset_2_columns = pd.DataFrame({ 
            "target" : target_attributes,
            "target_index": target_index
        })


        if ground_truth is not None and not skip_ground_truth_processing:
            ground_truth.columns = ['source', 'target']
            ground_truth = pd.merge(ground_truth, dataset_1_columns, on="source", how='left')
            ground_truth = pd.merge(ground_truth, dataset_2_columns, on='target', how='left')
            ground_truth = ground_truth.drop(columns=['source', 'target'])

        return dataset_1, dataset_2, ground_truth
    
        
    def load_schema(self,
            dataset_1: DataFrame, 
            dataset_2: DataFrame,
            ground_truth: DataFrame = None,
            skip_ground_truth_processing: bool = False
    ) -> tuple:
        
        dataset_1 = dataset_1.astype(str)
        dataset_2 = dataset_2.astype(str)

        source_attributes = dataset_1.columns
        target_attributes = dataset_2.columns

        source_index = range(len(source_attributes)) 

        target_index = range(len(target_attributes)) 


        dataset_1 = pd.DataFrame({
            'id' : source_index,
            'attributes': source_attributes,
        })

        dataset_2 = pd.DataFrame({
            'id' : target_index,
            'attributes': target_attributes,
        })


        dataset_1_columns = pd.DataFrame({
            'source': source_attributes,
            "source_index": source_index
        })

        dataset_2_columns = pd.DataFrame({ 
            "target" : target_attributes,
            "target_index": target_index
        })


        if ground_truth is not None and not skip_ground_truth_processing:
            ground_truth.columns = ['source', 'target']
            ground_truth = pd.merge(ground_truth, dataset_1_columns, on="source", how='left')
            ground_truth = pd.merge(ground_truth, dataset_2_columns, on='target', how='left')
            ground_truth = ground_truth.drop(columns=['source', 'target'])

        return dataset_1, dataset_2, ground_truth


import math

from pyjedai.datamodel import SpatialData, PYJEDAIFeature
from queue import PriorityQueue
from tqdm.auto import tqdm


class AbstractSpatialInitialization(PYJEDAIFeature):
    """Abstract class for the block building method
    """
    def __init__(self):
        super().__init__()
        return
    
    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True,
                 with_stats: bool = False
                ) -> any:
        # TODO itile

        return

    def stats(self, blocks: dict) -> None:
        # TODO itile

        return

class StandardSpatialInitialization(AbstractSpatialInitialization):
    def __init__(self,  budget:int, wScheme:str):
        super().__init__()
        self.wScheme = wScheme
        self.budget = budget

    def process(self, spatial_data:SpatialData, spatial_index:list, theta_x:float, theta_y:float): # reads target geometries on the fly
        self.source_geometries = spatial_data.source_geometries
        self.targetGeometries = spatial_data.targetGeometries
        self.spatial_index = spatial_index
        self.theta_x = theta_x
        self.theta_y = theta_y

        self.flag = [-1] * spatial_data.source_geometries_size
        self.freq = [-1] * spatial_data.source_geometries_size
        self.topKPairs = PriorityQueue(maxsize = self.budget + 1)

        minimumWeight = -1
        targetId = 0

        # TODO itile spatial.size2
        
        for targetGeom in self.targetGeometries:
            candidates = self.getCandidates(targetId, targetGeom)
            for candidateMatchId in candidates:
                if (self.validCandidate(candidateMatchId, targetGeom.envelope)):
                    weight = self.getWeight(candidateMatchId, targetGeom)
                    if (minimumWeight <= weight):
                        self.topKPairs.put((weight, candidateMatchId, targetId, targetGeom))
                        if (self.budget < self.topKPairs.qsize()):
                            minimumWeight = self.topKPairs.get()[0]
            targetId += 1

        # print("Total target geometries", targetId)
        return(self.topKPairs)

    def getCandidates(self, targetId, tEntity):
        candidates = set()

        envelope = tEntity.envelope.bounds
        maxX = math.ceil(envelope[2] / self.theta_x)
        maxY = math.ceil(envelope[3] / self.theta_y)
        minX = math.floor(envelope[0] / self.theta_x)
        minY = math.floor(envelope[1] / self.theta_y)

        for latIndex in range(minX, maxX):
            for longIndex in range(minY,maxY):
                for sourceId in self.spatial_index[latIndex][longIndex]:
                    if (self.flag[sourceId] == -1):
                        self.flag[sourceId] = targetId
                        self.freq[sourceId] = 0
                    self.freq[sourceId] += 1
                    candidates.add(sourceId)

        return candidates

    def validCandidate(self, candidateId, targetEnv):
        return self.source_geometries[candidateId].envelope.intersects(targetEnv)

    def getWeight(self,sourceId, tEntity) :
        commonBlocks = self.freq[sourceId]
        if self.wScheme == 'CF':
            return commonBlocks
        elif self.wScheme == 'JS_APPROX':
            return commonBlocks / (self.getNoOfBlocks(self.source_geometries[sourceId].envelope) + self.getNoOfBlocks(tEntity.envelope) - commonBlocks)
        elif self.wScheme == 'MBR':
            srcEnv = self.source_geometries[sourceId].envelope
            trgEnv = tEntity.envelope
            mbrIntersection = srcEnv.intersection(trgEnv)
            denominator = srcEnv.area + trgEnv.area - mbrIntersection.area
            if denominator == 0 :
                return 0
            return mbrIntersection.area/denominator
        return 1.0

    def getNoOfBlocks(self,envelope) :
        maxX = math.ceil(envelope[2] / self.theta_x)
        maxY = math.ceil(envelope[3] / self.theta_y)
        minX = math.floor(envelope[0] / self.theta_x)
        minY = math.floor(envelope[1] / self.theta_y)
        return (maxX - minX + 1) * (maxY - minY + 1)

    def _configuration(self) -> dict:
        """No configuration"""
        return {}



from tqdm.auto import tqdm
from pyjedai.datamodel import SpatialData, PYJEDAIFeature
 
from shapely.geometry import multipolygon
from collections import defaultdict
import math

class AbstractSpatialFiltering(PYJEDAIFeature):
    """Abstract class for the block building method
    """
    def __init__(self):
        super().__init__()

        return
    
    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True,
                 with_stats: bool = False
                ) -> any:
        # TODO itile

        return

    def stats(self, blocks: dict) -> None:
        # TODO itile

        return

class StandardSpatialFiltering(AbstractSpatialFiltering):
    """Abstract class for the block building method
    """
    def __init__(self):
        super().__init__()
        self.spatial_index = defaultdict(lambda: defaultdict(list))

        return

    def process(self, spatial_data:SpatialData)-> dict:
        self.geometries = spatial_data.source_geometries

        self.setThetas()
        self.indexSource()

        # return {"spatial_index":self.spatial_index,"self.theta_x":self.theta_x,"self.theta_y":self.theta_y}
        return (self.spatial_index, self.theta_x, self.theta_y)

    def setThetas(self):
        self.theta_x, self.theta_y = 0, 0
        for sEntity in self.geometries:
            if not isinstance(sEntity, multipolygon.MultiPolygon):
                print("Warning non geometry oject in filtering")
                continue

            envelope = sEntity.envelope.bounds
            self.theta_x += envelope[2] - envelope[0]
            self.theta_y += envelope[3] - envelope[1]

        if(len(self.geometries) != 0):
            self.theta_x /= len(self.geometries)
            self.theta_y /= len(self.geometries)
            # print("\nDimensions of Equigrid", self.theta_x,"and", self.theta_y)
        else:
            print("Error in setThetas(), division by zero")

    def indexSource(self) :
        geometryId = 0
        for sEntity in self.geometries:
            self.addToIndex(geometryId, sEntity.bounds)
            geometryId += 1

    def addToIndex(self, geometryId, envelope) :
        maxX = math.ceil(envelope[2] / self.theta_x)
        maxY = math.ceil(envelope[3] / self.theta_y)
        minX = math.floor(envelope[0] / self.theta_x)
        minY = math.floor(envelope[1] / self.theta_y)

        for latIndex in range(minX, maxX):
            for longIndex in range(minY, maxY):
                self.spatial_index[latIndex][longIndex].append(geometryId)

    def _configuration(self) -> dict:
        """No configuration"""
        return {}


"""Schema Matching methods
"""
import pandas as pd
import valentine
from valentine.algorithms.base_matcher import BaseMatcher
from valentine.algorithms.coma.coma import Coma
from valentine.algorithms.cupid.cupid_model import Cupid
from valentine.algorithms.distribution_based.distribution_based import DistributionBased
from valentine.algorithms.jaccard_distance.jaccard_distance import JaccardDistanceMatcher
from valentine.algorithms.similarity_flooding.similarity_flooding import SimilarityFlooding
import valentine.metrics as valentine_metrics
from pandas import DataFrame, concat

from ..datamodel import Block, SchemaData, PYJEDAIFeature
from ..evaluation import Evaluation
from abc import abstractmethod

class AbstractSchemaMatching(PYJEDAIFeature):
    """Abstract class for schema matching methods
    """

    @abstractmethod
    def evaluate(self,
                 prediction=None,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:
        pass

    @abstractmethod
    def _configuration(self) -> dict:
        pass

    @abstractmethod
    def stats(self) -> None:
        pass

    @abstractmethod
    def process(self,
                data: SchemaData,
                ) -> list:
        pass

    @abstractmethod
    def process_sm_weighted(self,
                            data: SchemaData):
        pass

    def __init__(self):
        super().__init__()


class ValentineMethodBuilder(PYJEDAIFeature):
    """Class to provide valentine matching methods
    """

    def evaluate(self,
                 prediction=None,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:
        pass

    def _configuration(self) -> dict:
        pass

    def __init__(self):
        super().__init__()

    @staticmethod
    def coma_matcher(max_n: int = 0,
                     use_instances: bool = False
                     ) -> Coma:
        return Coma(max_n, use_instances)

    @staticmethod
    def cupid_matcher(w_struct: float = 0.2,
                      leaf_w_struct: float = 0.2,
                      th_accept: float = 0.7
                      ) -> Cupid:
        return Cupid(w_struct, leaf_w_struct, th_accept)

    @staticmethod
    def distribution_based_matcher(threshold1: float = 0.15,
                                   threshold2: float = 0.15
                                   ) -> DistributionBased:
        return DistributionBased(threshold1, threshold2)

    @staticmethod
    def jaccard_distance_matcher(threshold_leven: float = 0.8) -> JaccardDistanceMatcher:
        return JaccardDistanceMatcher(threshold_leven)

    @staticmethod
    def similarity_flooding_mathcer(coeff_policy: str = "inverse_average",
                                    formula: str = "formula_c") -> SimilarityFlooding:
        return SimilarityFlooding(coeff_policy, formula)

class ValentineSchemaMatching(AbstractSchemaMatching):
    """Class for schema matching methods provided by Valentine
    """

    def __init__(self, matcher: BaseMatcher):
        super().__init__()
        self.data: SchemaData = None
        self.matcher: BaseMatcher = matcher
        self.matches = None
        self.top_columns: list = []

    def process(self,
                data: SchemaData,
                ) -> list:
        self.data = data
        df1 = self.data.dataset_1
        df2 = self.data.dataset_2
        self.matches = valentine.valentine_match(df1, df2, self.matcher)
        self.top_columns = [[x[0][1] for x in self.matches.keys()], [x[1][1] for x in self.matches.keys()]]
        return self.top_columns

    def process_sm_weighted(self, data: SchemaData):
        pass

    def print_matches(self):
        for match, sim in self.matches.items():
            print(match, " - ", sim)
        
    def get_matches(self) -> dict:
        return self.matches

    def evaluate(self,
                 prediction=None,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                                 "Data object has not been initialized with the ground-truth file")

        return valentine_metrics.all_metrics(self.matches, self.data.ground_truth)

    def _configuration(self) -> dict:
        pass

    def stats(self) -> None:
        pass




import pandas as pd
import pytest

from pyjedai.datamodel import Data
from pyjedai.vector_based_blocking import EmbeddingsNNBlockBuilding


@pytest.fixture(scope="module")
def data():
    # Mock data object
    return Data(
        dataset_1=pd.DataFrame({"attribute_1": ["entity1", "entity2"], "id": [1, 2]}),
        dataset_2=pd.DataFrame({"attribute_1": ["entity3", "entity4"], "id": [3, 4]}),
        attributes_1=["attribute_1"],
        attributes_2=["attribute_1"],
        id_column_name_1="id",
        id_column_name_2="id",
    )

def test_custom_vectorizer_standard_model_raises_error(data):
    instance = EmbeddingsNNBlockBuilding(vectorizer="word2vec")

    with pytest.raises(AttributeError):
        instance.build_blocks(data, custom_pretrained_model="word")

def test_custom_vectorizer_word_embeddings(data, mocker):
    instance = EmbeddingsNNBlockBuilding(vectorizer='custom')
    # mock the tokenizer and model loading
    assert instance.vectorizer == 'custom'
    mock_tokenizer = mocker.patch("pyjedai.vector_based_blocking.AutoTokenizer.from_pretrained")
    mock_model = mocker.patch("pyjedai.vector_based_blocking.AutoModel.from_pretrained")
    instance.build_blocks(data, custom_pretrained_model="word")
    # Check if the tokenizer and model were called once
    assert mock_tokenizer.call_count == 1
    assert mock_model.call_count == 1


def test_custom_vectorizer_sentence_embeddings(data, mocker):
    instance = EmbeddingsNNBlockBuilding(vectorizer='custom')
    assert instance.vectorizer == 'custom'
    mock_transformer = mocker.patch("pyjedai.vector_based_blocking.SentenceTransformer")

    # check if the transformer was called with the correct model
    instance.build_blocks(data, custom_pretrained_model="sentence")
    assert mock_transformer.call_count == 1
    assert mock_transformer.call_args[0][0] == "custom"